"""Experimental 'python -m chemfp' command driver

This code is undocumented and only lightly tested. The current goal is
to get feedback from friendly users.

"""
# Copyright (c) 2020 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See http://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

from __future__ import absolute_import

import sys
import argparse

import os

from . import commandline
from .commandline import cmdsupport

_commands = {}
_command_list = []

def command(command_name):
    def f(command_func):
        _commands[command_name] = command_func
        _command_list.append(command_name)
        return command_func
    return f

@command("simsearch")
def simsearch_command():
    from .commandline import simsearch
    simsearch.run()

@command("rdkit2fps")
def rdkit2fps_command():
    commandline.run_rdkit2fps()

@command("oe2fps")
def oe2fps_command():
    commandline.run_oe2fps()

@command("ob2fps")
def ob2fps_command():
    commandline.run_ob2fps()

@command("sdf2fps")
def sdf2fps_command():
    from .commandline import sdf2fps
    sdf2fps.run()

@command("fpcat")
def fpcat_command():
    from .commandline import fpcat
    fpcat.run()
#### 

# Some bonus commands that aren't (yet?) command-line tools

def _add_in_out_args(parser):
    parser.add_argument(
        "--in", metavar="FORMAT", dest="input_format",
        choices=("fps", "fps.gz", "fps.zst", "fpb", "fpb.gz", "fpb.zst", "flush"),
        help="input fingerprint format. One of fps or fpb (with optional gz or zst compression), or flush. (default guesses from filename or is fps)")
    
    parser.add_argument(
        "-o", "--output", metavar="FILENAME",
        help="save the fingerprints to FILENAME (default=stdout)")
    parser.add_argument(
        "--out", metavar="FORMAT", dest="output_format", choices=("fps", "fps.gz", "fps.zst", "fpb", "flush"),
        help="output fingerprint format. One of fps, fps.gz, fps.zst, fpb, or flush. (default guesses from output filename, or is 'fps')")

    parser.add_argument("filename", nargs="?",
                            help = "fingerprint filename. If not given, read from stdin.")
    
@command("first")
def first_command():
    import itertools
    from . import open as open_fingerprints
    from . import open_fingerprint_writer
    from . import types
    parser = argparse.ArgumentParser(
        prog = "first",
        description = "select up to the the first N fingerprints")
    parser.add_argument("-n", "--n", type=types.nonnegative_int, default=10,
                            help="number of fingerprints")
    _add_in_out_args(parser)

    args = parser.parse_args()

    reader = open_fingerprints(args.filename, format=args.input_format)
    if hasattr(reader, "arena"):
        subset = reader[:args.n]
        subset.save(args.output, format=args.output_format)
        return

    with open_fingerprint_writer(args.output, format=args.output_format,
                                     metadata=reader.metadata) as writer:
        writer.write_fingerprints(itertools.islice(reader, 0, args.n))
    
    
@command("last")
def last_command():
    import collections
    from . import open as open_fingerprints
    from . import open_fingerprint_writer
    from . import types
    parser = argparse.ArgumentParser(
        prog = "last",
        description = "select up to the the last N fingerprints")
    parser.add_argument("-n", "--n", type=types.nonnegative_int, default=10,
                            help="number of fingerprints")
    _add_in_out_args(parser)

    args = parser.parse_args()

    reader = open_fingerprints(args.filename, format=args.input_format)
    if hasattr(reader, "arena"):
        subset = reader[:args.n]
        subset.save(args.output, format=args.output_format)
        return

    with open_fingerprint_writer(args.output, format=args.output_format,
                                     metadata=reader.metadata) as writer:
        writer.write_fingerprints(collections.deque(reader, maxlen=args.n))


@command("sample")
def sample_command():
    from . import load_fingerprints
    from . import types
    parser = argparse.ArgumentParser(
        prog = "sample",
        description = "sample N fingerprints")
    parser.add_argument("-n", "--n", type=types.nonnegative_int, default=10,
                            help="number of fingerprints to sample (default: 10)")
    parser.add_argument("--upper-bound", "-u", action="store_true",
                            help="use all the fingerprints if there are fewer than 'n' of them")
    parser.add_argument("--no-reorder", action="store_true",
                            help="do not reorder the sampled fingerprints by popcount")
    parser.add_argument("--seed", type=int, default=None,
                            help="RNG seed to use for sampling (default: use the system RNG)")
    _add_in_out_args(parser)
    args = parser.parse_args()

    arena = load_fingerprints(args.filename, format=args.input_format, reorder=False)

    n = args.n
    if args.upper_bound:
        n = min(n, len(arena))
    elif n > len(arena):
        sys.stderr.write("Sample size %d is larger than the number of fingerprints (%d).\n" % (
            n, len(arena)))
        sys.stderr.write("Use a smaller value for --n or specify --upper-bound to use all available fingerprints.\n")
        raise SystemExit(1)
    sample = arena.sample(n, reorder=not args.no_reorder, rng=args.seed)
    sample.save(args.output, format=args.output_format)
    

# @command("train_test_split")
#### 
@command("fpb_dump")
def fpb_dump_command():
    from . import fpb_io
    fpb_io.main()
    
#### 
@command("report")
def report_command():
    parser = argparse.ArgumentParser(
        description = "report similarity search implementation details")
    args = parser.parse_args()
    from . import bitops
    bitops.use_environment_variables()
    bitops.print_report()
    
@command("toolkits")
def toolkits_command():
    parser = argparse.ArgumentParser(
        description = "show cheminformatics toolkit availability")
    args = parser.parse_args()
    
    def _software():
        return software.partition("/")[2]
    
    try:
        import openbabel
    except ImportError as err:
        print("openbabel - not installed: %s" % (err,))
    else:
        from .openbabel_toolkit import software
        print("openbabel - version %s installed" % (_software(),))

    try:
        import rdkit
    except ImportError as err:
        print("rdkit - not installed: %s" % (err,))
    else:
        from .rdkit_toolkit import software
        print("rdkit - version %s installed" % (_software(),))

    try:
        import openeye
    except ImportError as err:
        print("oechem - not installed: %s" % (err,))
    else:
        from .openeye_toolkit import software
        
        from openeye import oechem, oegraphsim
        if not oechem.OEChemIsLicensed():
            print("oechem - version %s installed but OEChem is not licensed" % (_software(),))
        elif not oegraphsim.OEGraphSimIsLicensed():
            print("oechem - version %s installed and OEChem licensed but OEGraphSim is not licensed" % (_software(),))
        else:
            print("oechem - version %s installed and both OEChem and OEGraphSim are licensed" % (_software(),))

@command("license")
def license_command():
    parser = argparse.ArgumentParser(
        description = "show the chemfp license status")
    args = parser.parse_args()
        
    try:
        cmdsupport.license_check()
    except SystemExit:
        pass

###

epilog = """
Description of the available commands:    
 o simsearch: Tanimoto or Tversky fingerprint similarity search
 o rdkit2fps: use RDKit to generate fingerprints from a structure file
 o oe2fps: use OpenEye's OEChem and OEGraphSim to generate fingerprints
     from a structure file"
 o ob2fps: use Open Babel to generate fingerprints from a structure file
 o sdf2fps: extract fingerprints from a tag of records in an SD file
 o fpcat: convert between fingerprint file formats
 o first: select the first 'n' fingerprints
 o last: select the last 'n' fingerprints
 o sample: select 'n' fingeprints at random 
 o fpb_dump: display low-level dump of the contents of an FPB file
 o report: report similarity search implementation details
 o toolkits: show cheminformatics toolkit availability
 o license: report on the chemfp license status

This interface is EXPERIMENTAL and UNSUPPORTED. Let me know if it's useful
or if there are other tools you would like.
"""
def _check_epilog():
    for line in epilog.splitlines():
        pass
    
parser = argparse.ArgumentParser(
    prog = "chemfp",
    usage = "python -m chemfp [-h] [-S] command [args]",
    description = "chemfp tools for working with cheminformatics fingerprints",
    formatter_class = argparse.RawDescriptionHelpFormatter,
    epilog = epilog)
parser.add_argument("--allow-signals", "-S", action = "store_true",
                        help = "let Python handle SIGPIPE and SIGITR signals instead of exiting")

def _get_command_list_str():
    terms = []
    for name in _command_list:
        terms.append(repr(name))
        terms.append(", ")
    assert len(terms) > 4
    del terms[-1]
    terms[-2] = " or "
    return "".join(terms)

_command_list_str = _get_command_list_str()
parser.add_argument("command", nargs="?",
                        help = "one of: " + _command_list_str)

parser.add_argument("args", nargs=argparse.REMAINDER,
                        help = "arguments passed to the subcommand; use command --help for details")

def main(argv=None):
    args = parser.parse_args(argv)
    command = args.command
    if command is None:
        parser.print_usage()
        print(epilog)
        return

    command_func = _commands.get(command, None)
    if command_func is None:
        parser.error("unsupported command: %r (Must be one of %s" % (
            command, _command_list_str))
        
    sys.argv[:] = [command] + args.args
    
    if args.allow_signals:
        def run(f):
            return f()
        cmdsupport.run = run

    command_func()

if __name__ == "__main__":
    main()
